@@ -12,7 +12,19 @@ except ImportError:
     import urllib.parse as urlparse
 
 if sys.version_info >= (3, 0, 0):
-    basestring = str
+    from io import BytesIO
+else:
+    try:
+        from cStringIO import StringIO as BytesIO
+    except ImportError:
+        from StringIO import StringIO as BytesIO
+
+
+binary_type = bytes
+if sys.version_info >= (3, 0, 0):
+    text_type = str
+else:
+    text_type = unicode  # noqa
 
 
 class Headers(object):
@@ -27,14 +39,13 @@ class Headers(object):
 
 
 def response(status_code=200, content='', headers=None, reason=None, elapsed=0,
-             request=None):
+             request=None, stream=False):
     res = requests.Response()
     res.status_code = status_code
-    if isinstance(content, dict):
-        if sys.version_info[0] == 3:
-            content = bytes(json.dumps(content), 'utf-8')
-        else:
-            content = json.dumps(content)
+    if isinstance(content, (dict, list)):
+        content = json.dumps(content).encode('utf-8')
+    if isinstance(content, text_type):
+        content = content.encode('utf-8')
     res._content = content
     res._content_consumed = content
     res.headers = structures.CaseInsensitiveDict(headers or {})
@@ -48,6 +59,10 @@ def response(status_code=200, content='', headers=None, reason=None, elapsed=0,
     if 'set-cookie' in res.headers:
         res.cookies.extract_cookies(cookies.MockResponse(Headers(res)),
                                     cookies.MockRequest(request))
+    if stream:
+        res.raw = BytesIO(content)
+    else:
+        res.raw = BytesIO(b'')
 
     # normally this closes the underlying connection,
     #  but we have nothing to free.
@@ -63,7 +78,7 @@ def all_requests(func):
     return inner
 
 
-def urlmatch(scheme=None, netloc=None, path=None, method=None):
+def urlmatch(scheme=None, netloc=None, path=None, method=None, query=None):
     def decorator(func):
         @wraps(func)
         def inner(self_or_url, url_or_request, *args, **kwargs):
@@ -79,6 +94,8 @@ def urlmatch(scheme=None, netloc=None, path=None, method=None):
                 return
             if path is not None and not re.match(path, url.path):
                 return
+            if query is not None and not re.match(query, url.query):
+                return
             if method is not None and method.upper() != request.method:
                 return
             return func(self_or_url, url_or_request, *args, **kwargs)
@@ -86,6 +103,38 @@ def urlmatch(scheme=None, netloc=None, path=None, method=None):
     return decorator
 
 
+def handler_init_call(handler):
+    setattr(handler, 'call', {
+        'count': 0,
+        'called': False
+    })
+
+
+def handler_clean_call(handler):
+    if hasattr(handler, 'call'):
+        handler.call.update({
+            'count': 0,
+            'called': False
+        })
+
+
+def handler_called(handler, *args, **kwargs):
+    try:
+        return handler(*args, **kwargs)
+    finally:
+        handler.call['count'] += 1
+        handler.call['called'] = True
+
+
+def remember_called(func):
+    handler_init_call(func)
+
+    @wraps(func)
+    def inner(*args, **kwargs):
+        return handler_called(func, *args, **kwargs)
+    return inner
+
+
 def first_of(handlers, *args, **kwargs):
     for handler in handlers:
         res = handler(*args, **kwargs)
@@ -104,9 +153,14 @@ class HTTMock(object):
 
     def __enter__(self):
         self._real_session_send = requests.Session.send
+        self._real_session_prepare_request = requests.Session.prepare_request
+
+        for handler in self.handlers:
+            handler_clean_call(handler)
 
         def _fake_send(session, request, **kwargs):
-            response = self.intercept(request)
+            response = self.intercept(request, **kwargs)
+
             if isinstance(response, requests.Response):
                 # this is pasted from requests to handle redirects properly:
                 kwargs.setdefault('stream', session.stream)
@@ -136,18 +190,35 @@ class HTTMock(object):
                     history.insert(0, response)
                     response = history.pop()
                     response.history = tuple(history)
+
+                session.cookies = response.cookies
+
                 return response
 
             return self._real_session_send(session, request, **kwargs)
+
+        def _fake_prepare_request(session, request):
+            """
+            Fake this method so the `PreparedRequest` objects contains
+            an attribute `original` of the original request.
+            """
+            prep = self._real_session_prepare_request(session, request)
+            prep.original = request
+            return prep
+
         requests.Session.send = _fake_send
+        requests.Session.prepare_request = _fake_prepare_request
+
         return self
 
     def __exit__(self, exc_type, exc_val, exc_tb):
         requests.Session.send = self._real_session_send
+        requests.Session.prepare_request = self._real_session_prepare_request
 
-    def intercept(self, request):
+    def intercept(self, request, **kwargs):
         url = urlparse.urlsplit(request.url)
         res = first_of(self.handlers, url, request)
+
         if isinstance(res, requests.Response):
             return res
         elif isinstance(res, dict):
@@ -156,14 +227,15 @@ class HTTMock(object):
                             res.get('headers'),
                             res.get('reason'),
                             res.get('elapsed', 0),
-                            request)
-        elif isinstance(res, basestring):
-            return response(content=res)
+                            request,
+                            stream=kwargs.get('stream', False))
+        elif isinstance(res, (text_type, binary_type)):
+            return response(content=res, stream=kwargs.get('stream', False))
         elif res is None:
             return None
         else:
             raise TypeError(
-                "Dont know how to handle response of type {}".format(type(res)))
+                "Dont know how to handle response of type {0}".format(type(res)))
 
 
 def with_httmock(*handlers):
